###############################################################################
# David E. Konerding, LBNL
# See LBNLCopyright for copyright notice!
###############################################################################


""" Common functionality used by RFT clients """

import urlparse, sys, exceptions, os, optparse
from twisted.python import log
from twisted.internet import reactor
from pyGridWare.utility.scripts.client import \
    GetBasicOptParser, GetPortKWArgs, SetUp
    

from ZSI.parse import ParsedSoap
from twisted.python import log
import sys, copy, time, traceback
from twisted.internet import reactor
from pyGridWare.utility.scripts.client import \
    GetOptParser, GetPortKWArgs, SetUp, GetConfigParser
from pyGridWare.wsrf.notification.NotificationConsumer import NotificationConsumer, NotificationConsumerService
from pyGridWare.wsrf.notification import SubscribeUtility

## TODO: change to pyGridWare
from pyGridWare.generated.stubs import reliable_transfer_factory_service as RFT_FACTORY_SERVICE
from pyGridWare.generated.stubs import reliable_transfer_service as RFT_SERVICE

## this import might not be necessary
from pyGridWare.generated.types import reliable_transfer_service as RFT_SERVICE_TYPES
from pyGridWare.common.delegationlib import delegate_cred

from ZSI.TC import _get_type_definition as TED
from ZSI.TC import _get_global_element_declaration as GED
TransferType_Def = TED('http://www.globus.org/namespaces/2004/10/rft', "TransferType")(None)
RFTOptionsType_Def = TED('http://www.globus.org/namespaces/2004/10/rft', "RFTOptionsType")(None)
TransferRequestType_Def = TED('http://www.globus.org/namespaces/2004/10/rft', "TransferRequestType")(None)
OverallStatusType_Def = TED('http://www.globus.org/namespaces/2004/10/rft', "OverallStatus")
RequestStatusType_Def = TED('http://www.globus.org/namespaces/2004/10/rft', "RequestStatusType")
OverallStatusProperty_Dec = GED('http://www.globus.org/namespaces/2004/10/rft', "OverallStatusProperty")
RequestStatusProperty_Dec = GED('http://www.globus.org/namespaces/2004/10/rft', "RequestStatusProperty")
transferRequest_Dec = GED('http://www.globus.org/namespaces/2004/10/rft', "transferRequest")
CreateReliableFileTransferInputMessage = GED("http://www.globus.org/namespaces/2004/10/rft","createReliableFileTransferRequest").pyclass
SetTerminationTime_Dec = GED("http://docs.oasis-open.org/wsrf/2004/06/wsrf-WS-ResourceLifetime-1.2-draft-01.xsd", "SetTerminationTime").pyclass


class Holder:
    """ Empty Holder class used to contain a transfer request """
    pass


## Class that handles incoming notifications (from the server).
class RFTNotificationConsumerService(NotificationConsumerService):

    """ RFTNotificationConsumerService implements the
    NotificationConsumerService interface for the RFT client.  It
    responds to the notifications that the RFT service provides """

    def __init__(self):
        self._StatusCallback = None
        
    def wsa_Notify(self, ps, address, **kw):
        request, response = NotificationConsumerService.wsa_Notify(self, ps, address)
        pyobj = request._NotificationMessage[0]._Message
        value = pyobj._ResourcePropertyValueChangeNotification._NewValue._any

## 	print "Notify", value
##         print dir(value)
##         print "value.typecode=", value.typecode
        ## exit the program when we receive the first OverallStatus message.

        if self._StatusCallback:
            self._StatusCallback(value)
##         if isinstance(value.typecode, OverallStatusType_Def) and self._OverallStatus_callback:
##             self._OverallStatus_callback(value)
##         elif isinstance(value.typecode, RequestStatus_Def) and self._RequestStatus_callback:
##             self._RequestStatus_callback(value)

        return response, None



class RFT:

    """ Initialize RFT class to transfer a single file from source to
    destination.

    Contact the delegation service, delegate our credentials to
    it, then set up a transfer and wait until it completes.
    
    options - a class holding pyGridWare connection options (will
    be passed to GetPortKWArgs)

    source - URL of the source file to be transferred
    
    destination - URL of the destination file to be transferred

    status_callback - optional status callback that can be used by users of RFT to get status updates
    
    TODO: add an AddTransfer option that is used to populate the list of transfers.
    """
    def __init__(self, options, transferPairs, status_callback=None):

        self._options = options
        self._transferPairs = transferPairs
        if not status_callback:
            self._status_callback = self.StatusCallback
        else:
            self._status_callback = status_callback

        """ Properties: initialized to None and set to appropriate
        values during method calls """

        ## EPR of delegated credential
        self._delEPR = None
        ## EPR of RFT
        self._rftEPR = None
        ## Instance of notification consumer class
        self._consumer = None

        self._status = []

    def Transfer(self):

        """ Set up a transfer (by delegating the user credential,
        creating an RFT resource, creating a notification consumer
        instance, subscribing to the status RPs, and initiating the
        transfer)

        """

        ## support multiple transfers by re-using existing delegation
        ## EPR, RFT EPR (not sure this is allowed)
        if not self._rftEPR:
            if not self._delEPR:
                self.DelegateCredential()
                
            ## Connect to the RFT factory and get an RFT resource
            self.CreateRFTResource()

        ## Setup the RFT resource
        self.SetupRFT()

        ## Subscribe to notifications
        if not self._consumer:
            self.SetupNC()
            self.Subscribe('OverallStatusProperty')
            self.Subscribe('RequestStatusProperty')

        self.DoTransfer()
    def DelegateCredential(self):

        """ Convenience method to delegate credential """
        
        ## make a copy of the options, so we can set the URL to the DFS on the same server
        o = copy.deepcopy(self._options)
        o.url = urlparse.urljoin(self._options.url, "DelegationFactoryService")
        ## only needed if using secure message
        # o.m2SecureConv = o.url + "-authService"

        try:
            kw = GetPortKWArgs(o)
        except proxylib.InvalidProxyException:
            print "Cannot contact service with an invalid proxy"
            
            ## This support works ok for client applications but is
            ## not appropriate in a service environment.  We should be
            ## using deferreds and errbacks.
            
            reactor.stop()
            return -1

        ## Always do full delegation
        kw['full'] = True
        self._options.limited = False

        self._delEPR = delegate_cred(**kw)

        ## ## Test code to establish that we can deserialize an EPR of an
        ## ## existing delegation
        ##     input = open("del.epr.pygridware").read()
        ##     ps = ParsedSoap(input, envelope=False)
        ##     EndpointReferenceType= TED("http://schemas.xmlsoap.org/ws/2004/03/addressing", "EndpointReferenceType")
        ##     delEPR = ps.Parse(EndpointReferenceType('DelegatedEPR'))


    ## Create an RFT context and populate it with a single transfer
    def CreateRFTResource(self):

        """ Contact the RFT Factory Service and get an RFT Resource.  Set various RFT transfer options
        and set up the list of files to transfer """

        transfers = []

        for pair in self._transferPairs:
            transfer = TransferType_Def.pyclass()
            transfer.SourceUrl = pair[0]
            transfer.DestinationUrl = pair[1]

            ## TODO: better control over all the transfer options
            transfer.SourceSubjectName = None
            transfer.DestinationSubjectName = None

            transfers.append(transfer)

        ## TODO: better control over all the RFT options
        rftOptions = RFTOptionsType_Def.pyclass()
        rftOptions.Binary = True

        # haven't been able to get this to work
        ## request = transferRequest_Dec.pyclass()
        request = Holder()
        request._rftOptions = rftOptions
        ## set up a single transfer
        request._transfer = transfers
        
        self._delEPR.typecode = None
        request._transferCredentialEndpoint = self._delEPR

        locator = RFT_FACTORY_SERVICE.ReliableFileTransferFactoryServiceLocator()
        kw = GetPortKWArgs(self._options)
        kw['full'] = not self._options.limited
        port = locator.getReliableFileTransferFactoryPortType(portAddress=self._options.url, **kw)
        request2 = RFT_FACTORY_SERVICE.createReliableFileTransferInputMessage()
        request2._transferRequest = request

        ## TODO: remove arbitrary constants
        ## TODO: figure out which of these is right
        request2._InitialTerminationTime = time.gmtime(time.time()+10000)
        request2.InitialTerminationTime = time.gmtime(time.time()+10000)
        
        response = port.createReliableFileTransfer(request2)
        self._rftEPR = response._reliableTransferEPR


    def SetupNC(self):

        """ Set up an RFT Notification consumer service, and register
        our status callback method with it.  Then start the service.
        """
        
        self._consumer = NotificationConsumer(notificationClass=RFTNotificationConsumerService)
        self._consumer.consumer._StatusCallback = self._status_callback
        self._consumer.start()

    def Subscribe(self, name):

        """ Subscribe to a resource property in the RFT service """
        
        SCHEMA_NS = "http://www.globus.org/namespaces/2004/10/rft"
        request = RFT_SERVICE.SubscribeRequest()
        SubscribeUtility.SimpleTopicExpression(request,
                                               self._consumer, 
                                               SCHEMA_NS,
                                               name=name)
        response = self._rftPort.Subscribe(request)

    def SetupRFT(self):

        """ Setup the RFT resource: basically, just get the port corresponding to the RFT resource EPR """
        o = copy.deepcopy(self._options)
        ## point URL at the RFT service
        o.url = urlparse.urljoin(self._options.url, "ReliableFileTransferService")
        kw = GetPortKWArgs(o)
        locator = RFT_SERVICE.ReliableFileTransferServiceLocator()
        self._rftPort = locator.getReliableFileTransferPortType(endPointReference=self._rftEPR, **kw)

    def DoTransfer(self):
        termTime = time.gmtime(time.time()+10000)
        request3 = RFT_SERVICE.SetTerminationTimeRequest()
        request3.RequestedTerminationTime = termTime
        response = self._rftPort.SetTerminationTime(request3)
        request4 = RFT_SERVICE.StartInputMessage()
        response = self._rftPort.start(request4)

    def StatusCallback(self, value):
        if isinstance(value.typecode, OverallStatusType_Def):
            for item in '_transfersActive', '_transfersCancelled', '_transfersFailed', '_transfersFinished', '_transfersPending', '_transfersRestarted':
                print item, ":", getattr(value, item)
            fault = value.Fault
            if fault:
                print 'Fault description:', value.Fault.RftTransferFaultType.Description
                print "Fault ErrorCode:", value.Fault.RftTransferFaultType.ErrorCode._text

            if value._transfersFailed + value._transfersFinished + value._transfersCancelled == len(self._transferPairs):
                reactor.stop()


        elif isinstance(value.typecode, RequestStatusType_Def):
            print "_requestStatus:", value._requestStatus
