###########################################################################
# Joshua R. Boverhof, LBNL
# See Copyright for copyright notice!
###########################################################################
import time
from zope.interface import implements
from pyGridWare.wsrf.faults.PropertiesFaults import ResourceUnknownFault
from interfaces import IResourceHome


class SimpleResourceHome:
    """Simple Resource Home, in-memory IResourceHome implementation.
    """
    implements(IResourceHome)
    resourceID = 1000

    def __init__(self):
        self.resources = {}
        self.avatarId = None

    def add(self, ctx, key=None):
        if key: ctx.resourceID = key
        key = ctx.resourceID
        try: 
            if int(key) > SimpleResourceHome.resourceID:
                SimpleResourceHome.resourceID = int(key)
                self.create()
        except ValueError:
            # allow anything
            #SimpleResourceHome.resourceID = str(key)
            pass
            
        self.resources[str(key)] = ctx

    def create(self):
        key = SimpleResourceHome.resourceID 
        SimpleResourceHome.resourceID += 1
        return key

    def delete(self, key):
        del self.resources[str(key)]

    def get(self, key):
        ctx = self.resources.get(str(key))
        if ctx is None:
            raise ResourceUnknownFault(\
                    timestamp=time.gmtime(),
                    originator=None,
                    errorcode=None,
                    description=None,
                    faultcause=None
            )
        return ctx
    
    def getResources(self):
        return self.resources.values()
    
    def setTerminationTime(self, key, terminationTime):
        self.resources[str(key)].TerminationTime = terminationTime
        
    def setAvatarId(self, avatarId):
        """set avatarId
        """
        self.avatarId = avatarId
        
    def getAvatarId(self):
        """get avatarId
        """
        if self.avatarId is None:
            raise RuntimeError, 'avatarId has not been set.'
        return self.avatarId
    


