############################################################################
# Joshua R. Boverhof, LBNL
# See Copyright for copyright notice!
###########################################################################
import time, weakref
from twisted.internet import reactor
from twisted.python import log

from ResourceContext import ScheduledResourceTerminationContext


class ResourceSweeper:
    def __init__(self, delay=5.0):
        self.__homes = []
        self.delay = delay
#        self.resourceHomes = resourceHomes
        self.delayedCall = reactor.callLater(self.delay, self.sweep)

    def getSweeperDelay(self):
        return self._delay
    def setSweeperDelay(self, delay):
        self._delay = delay
        assert self.delayedCall.active(), "Resource Sweeper delayedCall not active"
        self.delayedCall.reset(delay)
    delay = property(getSweeperDelay, setSweeperDelay, None, "Sweeper Delay in Seconds")

#    def getResourceHome(self):
#        return self._resourceHome()
#    def setResourceHome(self, home):
#        self._resourceHome = weakref.ref(home)
#    resourceHome = property(getResourceHome, setResourceHome, None, "Resource Home")

    def addResourceHome(self, home):
        self.__homes.append(weakref.ref(home))

    def sweep(self):
        currentTime = self.delayedCall.getTime()
        log.msg("==> Delayed Call Sweep -- %.2f" %currentTime, debug=True)
        
        for home in map(lambda w: w(), self.__homes):
            log.msg('resource home "%s"' %home, debug=True)
            
            for resource in home.getResources():
                log.msg('resource "%s"' %resource, debug=True)
                
                if isinstance(resource, 
                      ScheduledResourceTerminationContext) is False:
                    continue
                
                resource.CurrentTime = currentTime
                terminationTime = resource.TerminationTime
                if terminationTime < currentTime:
                    key = resource.getResourceID()
                    log.msg("* Sweep terminate %s" %key, debug=True)
                    home.delete(key)
                
        self.delayedCall = reactor.callLater(self.delay, self.sweep)
        
        

