############################################################################
# Joshua R. Boverhof, LBNL
# See Copyright for copyright notice!
# Generated by setup.py
###########################################################################
import base64, time, sys, traceback
from M2Crypto import BIO, X509, RSA
from zope.interface import classProvides, implements, Interface

import twisted
from twisted.python import log

from ZSI import _get_element_nsuri_name
from ZSI.address import Address
from ZSI.schema import GED
from ZSI.wstools.Namespaces import WSA200403 as WSA

from ZSI.twisted.WSresource import WSResource as _WSResource,\
    CallbackChainInterface, DataHandler, DefaultHandlerChain, WSAddressHandler,\
    HandlerChainInterface

from pyGridWare.addressing.AddressingUtils import AddressingUtils
from pyGridWare.utility.http import GetURLFromRequest

from pyGridWare.generated.services.delegation_factory_service.DelegationFactoryService import\
    DelegationFactoryServiceWSRF as WSRFService, DelegationFactoryPortTypeContext
from pyGridWare.generated.services.stubs.delegation_factory_service.DelegationFactoryService import\
    DelegationFactoryService as STUB
from pyGridWare.generated.resource.delegation_factory_service.DelegationFactoryService import\
    DelegationFactoryPortTypeContext
from pyGridWare.wsrf.faults.PropertiesFaults import\
    ResourceUnknownFault, InvalidResourcePropertyQNameFault


"""WARNING: Service Description Violates Basic Profile

Operations RequestSecurityToken & RequestSecurityToken2 violate WS-I, so
a hack is required to distinquish between these operations.

http://www.ws-i.org/Profiles/BasicProfile-1.0-2004-04-16.html#refinement16645448

5.6.7 Wire Signatures for Operations
An endpoint that supports multiple operations must unambiguously identify the operation being invoked based on the input message that it receives. This is only possible if all the operations specified in the wsdl:binding associated with an endpoint have a unique wire signature.
R2710 The operations in a wsdl:binding in a DESCRIPTION MUST result in wire signatures that are different from one another.

The Profile defines the "wire signature" of an operation in a wsdl:binding to be the fully qualified name of the child element of the soap:Body of the SOAP input message it describes. For the case of an empty soap:Body this name is an empty string.
"""

EPR = GED(WSA.ADDRESS, "EndpointReference").pyclass
BST =  GED("http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd","BinarySecurityToken").pyclass
VALUE = "http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-x509-token-profile-1.0#X509PKIPathv1"
ENCODING_TYPE = "http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-soap-message-security-1.0#Base64Binary"


class WSAddressHandler2:
    """
    collisions in wsAction and root dicts, must use soapAction dict to determine
    operation name
    """
    implements(HandlerChainInterface)
    
    def processRequest(self, ps, **kw):
        # TODO: Clean this up
        resource = kw['resource']

        self.address = address = Address()
        address.parse(ps)
        action = address.getAction()
        if not action:
            raise WSActionException('No WS-Action specified in Request')

        request = kw['request']
        http_headers = request.getAllHeaders()
        soap_action = http_headers.get('soapaction')
        if soap_action and soap_action.strip('\'"') != action:
            raise WSActionException(\
                'SOAP Action("%s") must match WS-Action("%s") if specified.'\
                %(soap_action,action)
            )
            
        self.op_name = resource.soapAction[action]

        # Save WS-Address in ParsedSoap instance.
        ps.address = address
        return ps
        
    def processResponse(self, sw, **kw):
        if sw is None:
            self.address = None
            return
        
        request, resource = kw['request'], kw['resource']
        if isinstance(request, twisted.web.http.Request) is False:
            raise TypeError, '%s instance expected' %http.Request
                
        d = getattr(resource, 'wsAction', None)
        key = self.op_name
        if d is None or d.has_key(key) is False:
            raise WSActionNotSpecified,\
                'Error looking for key(%s) in wsAction dictionary(%s)' %(key, str(d))

        addressRsp = Address(action=d[key])
        if request.transport.TLS == 0:
            addressRsp.setResponseFromWSAddress(\
                 self.address, 'http://%s:%d%s' %(
                 request.host.host, request.host.port, request.path)
            )
        else:
            addressRsp.setResponseFromWSAddress(\
                 self.address, 'https://%s:%d%s' %(
                 request.host.host, request.host.port, request.path)
            )
            
        addressRsp.serialize(sw, typed=False)
        self.address = None
        return sw
    

class WSAddressCallbackHandler:
    classProvides(CallbackChainInterface)


class WSActionCallbackHandler:
    classProvides(CallbackChainInterface)

    @classmethod
    def processRequest(cls, ps, **kw):
        """Expects WS-Address Handler to have been called.

        invokes callback that should return a (request,response) tuple.
        representing the SOAP request and response respectively.
        ps -- ParsedSoap instance representing HTTP Body.
        request -- twisted.web.server.Request
        """
        resource = kw['resource']
        request = kw['request']
        #root = _get_element_nsuri_name(ps.body_root)
        #method = resource.root[root] 
        wsaction = ps.address.getAction()
        method = resource.soapAction[wsaction]

        if method is None:
            raise UnknownRequestException,\
                'failed to map request to a method: action(%s), root%s'\
                %(action,_get_element_nsuri_name(ps.body_root))
              
        req_pyobj,rsp_pyobj =\
            getattr(resource, method)(ps, ps.address, request=request)
        
        return rsp_pyobj


class ModifiedHandlerChainFactory:
    protocol = DefaultHandlerChain
    
    @classmethod
    def newInstance(cls):
        return cls.protocol(WSActionCallbackHandler, DataHandler, WSAddressHandler2())


class MyContext(DelegationFactoryPortTypeContext):
    def write(self, factory):
        from pyGsi import proxylib
        #cert = X509.load_cert(certfile)
        #stack = X509.X509_Stack() 
        #stack.push(cert)

        # new stack service cert
        cert = X509.load_cert( factory.certificateFileName )
        key = RSA.load_key( factory.privateKeyFileName )

        # reverse: push proxy cert, user cert
        stack = X509.X509_Stack() 
        token = self.properties.CertificateChain.Any
        dstack = X509.new_stack_from_der(base64.decodestring(token))
        while len(dstack):
            cert = dstack.pop()
            log.msg('cert: ', cert.get_subject(), debug=True)
            stack.push(cert)
 
        # Save Delegated Cred to a file 
        log.msg('save proxy file', debug=True)
        proxyfile = proxylib.create_write_file('/tmp/whatever.pem')
        bio = BIO.File(proxyfile)

        for cert in stack: 
            bio.write(cert.as_pem())

        key.save_key_bio(bio, cipher=None)
        bio.close() 


class DelegationFactoryService(WSRFService):
    """Delegate to Singleton resource
    """
    factory = ModifiedHandlerChainFactory

    def __init__(self, contextFactory):
        """Either specify contextFactory in constructor or pass it in
        by setting it on the server
        """
        WSRFService.__init__(self)
        self.contextFactory = contextFactory or self.server.contextFactory

    def getServiceCertificateChain(self):
        """return this services CertificateChain Resource Property.
        Not accessed thru resourceID
        """
        factory = self.contextFactory
        certfile = factory.certificateFileName
        stack = X509.X509_Stack() 
        cert = X509.load_cert(certfile)
        stack.push(cert)
        der_encoding = stack.as_der() 
        bst = BST(base64.encodestring(der_encoding))
        bst.set_attribute_ValueType(VALUE)
        bst.set_attribute_EncodingType(ENCODING_TYPE)
        decl = DelegationFactoryPortTypeContext().properties.declaration.pyclass()
        chain = decl.new_CertificateChain()
        chain.Any = bst
        return chain

    #@classmethod
    def NewResourceContextDelegationFactoryPortTypePort(self, request):
        """create a new resource context"""
        kw = {}
        url = request
        if not isinstance(request, basestring):
            url = GetURLFromRequest(request)
            kw["request"] = request

        ctx = MyContext()
        ctx.setAddress(url)
        home = ctx.getResourceHome()
        key = home.create(**kw)
        home.add(ctx, key, **kw)
        ctx.setResourceID(key)

        return ctx

    def wsa_GetResourceProperty(self, ps, address, request=None, **kw):
        """Grab ResourceContext if available, else create one that
        hands back the server's certificate chain.  If WS-Addressing
        information headers do not contain an EPR hand back the 
        server's certficate chain.  

        Does not conform with WS-Resource/WS-ResourceProperties, this is 
        a required operation but no EPR is used to reference the servers
        credential --> illegal.
        """
        log.msg("DelegationFactoryService.wsa_GetResourceProperty", debug=True)
        try:
            return WSRFService.wsa_GetResourceProperty(self, ps, address, **kw)
        except Exception, ex:
            ## TODO: Could be a bad access, with bad resourceID
            pass

        ## 
        ## No ResourceID, setup ResourceProperty using bindings
        ## 
        request,response = STUB.wsa_GetResourceProperty(self, ps, address)

        if request != ("http://www.globus.org/08/2004/delegationService", "CertificateChain"):
            raise InvalidResourcePropertyQNameFault(timestamp=time.gmtime(),
                    faultcause=('Unknown ResourceProperty: "%s"' %request))

        response.Any = [ self.getServiceCertificateChain() ]

        return request,response

    def wsa_RequestSecurityToken(self, ps, address, request=None, **kw):
        """factory method, returns an EPR in the response
        Arguments:
            ps -- ParsedSoap
            address -- WS-Addressing MIH
            request -- HTTPRequest
        """

        log.msg("DelegationFactoryService.wsa_RequestSecurityToken", debug=True)
        msg,response = WSRFService.wsa_RequestSecurityToken(self, ps, address)

        ctxattr = msg.get_attribute_Context()
        anyattr = msg.get_attribute_anyAttribute()

        if len(msg.Any) != 1: 
            # TODO: throw what's specified.
            raise RuntimeError, 'expecting a single binary security token'

        token = msg.Any[0]
        if token.typecode is not BST.typecode:
            # TODO: throw what's specified.
            raise RuntimeError, 'Unsupported Security Token'

        ctx = self.NewResourceContextDelegationFactoryPortTypePort(request)
        rp =  ctx.properties

        # Grab the underlying RP Document
        doc = rp.getResourcePropertiesDocument()
        chain = doc.new_CertificateChain()
        chain.Any = token

        # Set the RP in the resource properties wrapper
        rp.CertificateChain = chain

        # Write out the Delegated Credential
        ctx.write(self.contextFactory)

        #response.set_Attribute_Context('whatever')
        #response.set_Attribute_anyAttribute('whatever')

        # Create EPR and return in response
        epr = AddressingUtils.createEndpointReference(ctx, pyobj=EPR())
        response.Any = [epr]

        return msg,response

#    def wsa_RequestSecurityToken2(self, ps, address, dumpctx=None, **kw):
#        """never used...
#        """
#        log.msg("DelegationFactoryService.wsa_RequestSecurityToken2", debug=True)
#        #request,response = WSRFService.wsa_RequestSecurityToken2(self, ps, address)
#        raise RuntimeError, 'Not Supported'
#    
