############################################################################
# Joshua R. Boverhof, LBNL
# See Copyright for copyright notice!
# Generated by setup.py
###########################################################################
from datetime import datetime
from twisted.python import log, runtime

from pyGridWare.utility.http import GetURLFromRequest
from pyGridWare.addressing.AddressingUtils import AddressingUtils
from pyGridWare.generated.services.reliable_transfer_factory_service.ReliableFileTransferFactoryService\
    import ReliableFileTransferFactoryServiceWSRF as WSRFService
from ReliableFileTransferService import ReliableFileTransferService as RFTService


def LaunchTransfer(src, dst, epr, credential):
    """
    src -- source URL
    dst -- destination URL
    credential -- DelegationFactoryService Resource Property instance 
    epr -- DelegatedCredential EPR
    """
    log.msg('LaunchTransfer', debug=True)
    ctx = ReliableFileTransferFactoryService.home.GetResourceContext(epr, credential)
    log.msg("RESOURCE: ", ctx.properties.CertificateChain, debug=True)
    return


class ReliableFileTransferFactoryService(WSRFService):
    """
            <xsd:element name="ReliableFileTransferFactoryProperties">
                <xsd:complexType>
                    <xsd:sequence>
                        <xsd:element ref="tns:ActiveResourceInstances"/>
                        <xsd:element ref="tns:TotalNumberOfTransfers"/>
                        <xsd:element ref="tns:TotalNumberOfActiveTransfers"/>
                        <xsd:element ref="tns:TotalNumberOfBytesTransferred"/>
                        <xsd:element ref="tns:RFTFactoryStartTime"/>
                        <xsd:element ref="tns:DelegationServiceEPR"/>
                    </xsd:sequence>
                </xsd:complexType>
            </xsd:element>

    # GetResourceContext
    # NewResourceContextReliableFileTransferFactoryPortTypePort

    """

    def wsa_createReliableFileTransfer(self, ps, address, request=None, **kw):
        """RFT FactoryService creates RFT Service instances..
        Keyword Arguments:
           request -- HTTPRequest instance

        """
        log.msg('createReliableFileTransfer', debug=True)
        msg,response = WSRFService.wsa_createReliableFileTransfer(self, ps, address)
        #<xsd:element minOccurs="0" name="InitialTerminationTime" nillable="true" type="xsd:dateTime"/>
        if msg.InitialTerminationTime:
            log.msg('intial termination time: ', msg.InitialTerminationTime, debug=True)

        # 
        # Creates an Anonymous resource,
        #   need to use "RFTService.portal" in order to 
        #   create a secure resource.
        resource = RFTService.NewResourceContextReliableFileTransferPortTypePort(request)


        #resource.ActiveResourceInstances
        #resource.TotalNumberOfTransfers
        #resource.TotalNumberOfActiveTransfers
        #resource.TotalNumberOfBytesTransferred
        #resource.RFTFactoryStartTime
        #resource.DelegationServiceEPR

        response.CurrentTime = datetime.utcfromtimestamp(runtime.seconds()).timetuple()
        response.NewTerminationTime = (msg.InitialTerminationTime or 
            datetime.utcfromtimestamp(runtime.seconds()+60*60*24).timetuple())
        response.ReliableTransferEPR = AddressingUtils.createEndpointReference(resource)

        try:
            credential = request.credential
        except:
            raise RuntimeError, 'Must specify a credential!!'

        delete = msg.DeleteRequest
        if delete:
            self.processDeleteRequest(delete, response, resource, credential)
            return msg,response

        self.processTransferRequest(msg.TransferRequest, response, resource, 
            credential=credential)

        return msg,response
 
    def processDeleteRequest(self, delete, response, resource, credential):
        log.msg('DeleteRequest: ', delete, debug=True)
        self.processBaseRequestType(delete, response, resource)

    def processTransferRequest(self, trans, response, resource, credential):
        log.msg('TransferRequest: ', trans, debug=True)

        self.processBaseRequestType(trans, response, resource)

        options = trans.RftOptions
        if options:
            self.processRftOptions(options, response, resource)

        if trans.AllOrNone is not None:
            self.processAll(trans.AllOrNone, response, resource)

        for t in trans.Transfer:
            self.processTransfer(t, response, resource, credential)

    def processAll(self, allOrNone, response, resource):
        log.msg('allOrNone', debug=True)

    def processRftOptions(self, options, response, resource):
        """ Type representing various options one can set at transfer and 
        request level.ignoreFilePermErr when set will force RFT to ignore
        errors that occur while setting/getting file permissions."""
        log.msg('rftOptions', debug=True)

        #<!-- true=binary false=ascii -->
        if options.Binary is True:
            pass
        #!-- blockSize in bytes -->
        bytes = options.BlockSize
        #<!-- tcpBufferSize in bytes -->
        size = options.TcpBufferSize 
        if options.Notpt is True:
            pass
    
        streams = options.ParallelStreams
    
        #<!-- DataChannelAuthentication -->
        if options.Dcau is True:
            pass
    
        #<!-- By default host authz is performed grid subjectName to be used 
        #        on both source and dest gftp servers for authz-->
        options.SubjectName
        options.DestinationSubjectName 
        options.SourceSubjectName 
    
        #<!-- username to be used to perform transfer. By default username
        # from gridmap lookup is used -->
        options.UserName 
        if options.IgnoreFilePermErr is True:
            pass

    def processTransfer(self, trans, response, resource, credential):
        """Type representing a single transfer along with various options that
        can be set on a transfer. retry_time is time when a particular 
        transfer is retried."""
        log.msg('process transfer: ', trans, debug=True)
        if trans.TransferId is not None:
            pass

        trans.SourceUrl
        trans.DestinationUrl
        trans.Attempts #int
        trans.Retry_time  #long
        trans.RftOptions 
        trans.Size  #long

        # TransferStatusTypeEnumeration
        status = trans.Status
        if not status:
            LaunchTransfer(trans.SourceUrl,
                trans.DestinationUrl,
                resource.DelegationServiceEPR,
                credential,
            )
        elif status not in ["Finished", "Retrying", "Failed", 
            "Active", "Pending", "Cancelled", "Expanding", 
            "Expanding_Done", "Dirty",]:
            log.msg('bad status: ', trans.Status, debug=True)


    def processBaseRequestType(self, pyobj, response, resource):
        """Base Request Type from which TransferRequestType and DeleteRequest
        Type are derived. Concurrency is the number of transfers done simul-
        taneously. maxAttempts denotes the number of attempts after which
        RFT marks a transfer as failed. finishBy (not implemented yet)
        represents the time by which this request should be finished.
        transferCredentialEndpoint represents EPR of delegation resource
        that holds the delegated credential for this resource."""
        log.msg('process BaseRequestType', debug=True)
    
        pyobj.Concurrency
        pyobj.MaxAttempts
        pyobj.FinishBy
        resource.DelegationServiceEPR = pyobj.TransferCredentialEndpoint



