############################################################################
# Joshua R. Boverhof, LBNL
# See Copyright for copyright notice!
###########################################################################
import os, random, types
from twisted.python import log as _log 
try: 
    from gov.lbl.dsd.netlogger import nllite
except ImportError:
    nllite = None

"""
Utilities for instrumenting code with netlogger.
"""

def instrument(func):
    """@instrument: decorator for instrumenting bound and
    class methods.  static methods should not be used.
    If nllite is unavailable or netlogger is not on,
    return the func so no performance penalty.
    """
    try:
        if nllite is None or not int(os.environ.get('NETLOGGER_ON')): 
            return func
    except:
        return func

    def netlog_wrapper_func(self, *args, **kw):
        if type(self) in (type, types.ClassType):
            event = 'auto.%s.%s' %(self.__name__, func.__name__)
        else:
            event = 'auto.%s.%s' %(self.__class__.__name__, func.__name__)

        callid = random.random()
        _log.msg('%s.start' %event, netlog=True, callid=callid)
        v = func(self, *args, **kw)
        _log.msg('%s.end' %event, netlog=True, callid=callid)
        return v

    return netlog_wrapper_func


class NetLoggerObserver:
    """
    Log observer that writes to a netlogd

    @type timeFormat: C{str} or C{NoneType}
    @ivar timeFormat: If not C{None}, the format string passed to strftime().
    """
    timeFormat = None

    def __init__(self, url):
        self.logstrm = nllite.LogOutputStream(url)
        self.logstrm.setMeta(None,{'GID':nllite.get_gid(), # Grid JOb Identifier
                      'PROG':nllite.get_prog(), # Program
                      'HOST':nllite.get_host(), # Host
                      'Program':'pyGridWare',})

    def __del__(self): 
        self.logstrm.close()

    def _emit(self, eventDict):
        if not eventDict.get("netlog"):
            return

        event="".join([str(m) for m in eventDict["message"]])
        del eventDict["message"]
        del eventDict["netlog"]
        self.logstrm.write(event, eventDict)

    def start(self):
        _log.addObserver(self._emit)

    def stop(self):
        _log.removeObserver(self._emit)


def addNetLoggerObserver():
    if int(os.environ.get('NETLOGGER_ON')):
        _log.addObserver(NetLoggerObserver(os.environ.get('NETLOGGER_DEST'))._emit)


def startNetLogging():
    global netlogObserver
    if not netlogObserver: 
        netlogObserver = NetLoggerObserver()

    _log.startLoggingWithObserver(netlogObserver._emit)

netlogObserver = None
