############################################################################
# Joshua R. Boverhof, LBNL
# See LBNLCopyright for copyright notice!
###########################################################################
import os, string
from ConfigParser import ConfigParser

# twisted
#   application
from twisted.application.service import Application
#   policies
from twisted.protocols.policies import WrappingFactory
#   python
from twisted.python import log
from twisted.python.failure import Failure
#   web
from twisted.web import server
from twisted.web.server import Site
from twisted.web.resource import Resource
from twisted.web.static import File
from twisted.web.script import ResourceScript
from twisted.web.http import HTTPChannel, datetimeToString

# ZSI
from ZSI.address import Address

# pyGridWare
from pyGridWare.utility.gsi import get_calist, GSIServer, GSIServerContextFactory
from pyGridWare.utility.m2crypto import TLSProtocolSessionCachingWrapper as TLSProtocolWrapper
from pyGridWare.wsrf.notification.SubscriptionManager import SubscriptionManager
from pyGridWare.services.SubscriptionManager import SubscriptionManager as SMServiceBinding

# if using globus need chunked
from pyGridWare.utility.http import ChunkedHTTPChannel
Site.protocol = ChunkedHTTPChannel


###########################################################################
# SETUP
#
###########################################################################
os.putenv("OPENSSL_ALLOW_PROXY_CERTS", "1")
CP = ConfigParser()
CP.read('server-config.txt')
PORT = CP.getint('Location', 'port')
HOST = CP.get('Location', 'host')


def SetUpResourceTree(root):    
    core = Resource()
    root.putChild('wsrf', Resource())
    root.getStaticEntity('wsrf').putChild('core', core)
    wsrf = Resource()

    # Discover WS-Address URI
    addr = Address()
    addr.setUp()

    # Initialize singleton
    manager = SubscriptionManager.getInstance(SMServiceBinding,
                      {'wsAddressURI':addr.wsAddressURI,
                       'contextFactory':core.server.args[1]},)
    core.putChild('SubscriptionManagerService', manager)  
    
    # Path "wsrf/services": Drop in directory "services"
    dropins = File("services")
    dropins.ignoreExt(".rpy")
    dropins.processors = {".rpy": ResourceScript}
    root.getStaticEntity('wsrf').putChild('services', dropins)

    # 
    # initialize portal authz infrastructure.
    # 
    log.msg('initialize all RPY services', debug=True)
    for rpy in filter(lambda f: f.endswith('.rpy'), dropins.listdir()):
        try:
            child = dropins.getChild(rpy, None),
        except ImportError, ex:
            log.msg('could not initialize service: ', rpy, debug=True)
        else:
            log.msg('initialized service: ', rpy, debug=True)
            

    return root


###########################################################################
# Set up Application
#
###########################################################################
class MyRequest(server.Request):
    """Present for logging events 
    """    

    def process(self):
        log.msg("Process a request.", debug=True)
        
        # get site from channel
        self.site = self.channel.site

        # set various default headers
        self.setHeader('server', server.version)
        self.setHeader('date', datetimeToString())
        self.setHeader('content-type', "text/html")

        # Resource Identification
        self.prepath = []
        self.postpath = map(server.unquote, string.split(self.path[1:], '/'))
        try:
            resrc = self.site.getResourceFor(self)
            self.render(resrc)
        except:
            self.processingFailed(Failure())
            
        self.method
        self.uri
        self.clientproto
        log.msg("CLIENT: %s" %self.client, debug=True)
        log.msg("HOST:   %s" %self.host, debug=True)

    
    
class MyHTTPChannel(HTTPChannel):
    """Present for logging events 
    """    
    def connectionMade(self):
        HTTPChannel.connectionMade(self)
        log.msg("CONNECTION MADE", debug=True)

    def dataReceived(self, data):
        log.msg("DATA RECEIVED: %s" %self.requestFactory, debug=True)
        return HTTPChannel.dataReceived(self, data)


root = Resource()
site = Site(root)
#site.protocol = MyHTTPChannel
site.requestFactory = MyRequest

capath = os.path.join('etc', 'grid-security', 'certificates')
contextFactory = GSIServerContextFactory(\
                     privateKeyFileName=os.path.join('etc', 'grid-security', 'containerkey.pem'),
                     certificateFileName=os.path.join('etc', 'grid-security', 'containercert.pem'),
                     calist=get_calist(capath),
                     capath=capath)

root.__class__.server = port = GSIServer(PORT, site, contextFactory=contextFactory, interface=HOST)


SetUpResourceTree(root)
application = Application("WSRF-GSI-Container")
port.setServiceParent(application)

