############################################################################
# Joshua R. Boverhof, LBNL
# See LBNLCopyright for copyright notice!
###########################################################################
import os, socket, warnings
from ConfigParser import ConfigParser
from twisted.application import service, internet
from twisted.web.server import Site
from twisted.web.resource import Resource
from twisted.web import static, script
from ZSI.address import Address
#from ZSI.wstools.logging import setBasicLoggerDEBUG
#setBasicLoggerDEBUG()

import pyGridWare
from pyGridWare.plugin import setDefaultPlugin
#setDefaultPlugin(opt, CP.get(SEC_PLUGINS, opt))
from pyGridWare.wsrf.notification.SubscriptionManager import SubscriptionManager
from pyGridWare.services.SubscriptionManager import SubscriptionManager as SMServiceBinding

# if using globus need chunked
from pyGridWare.utility.http import ChunkedHTTPChannel
Site.protocol = ChunkedHTTPChannel

#from pyGridWare.utility.log import addNetLoggerObserver
#addNetLoggerObserver()


###########################################################################
# Grab configuration information
#
###########################################################################
CP = ConfigParser()
CP.read('server-config.txt')
PORT = CP.getint('Location', 'port')
HOST = CP.get('Location', 'host')


###########################################################################
# SetUpResourceTree
#
###########################################################################
def SetUpResourceTree(root):
    # Discover WS-Address URI
    addr = Address()
    addr.setUp()
    
    # Path "wsrf/core"
    core = Resource()
    root.putChild('wsrf', Resource())
    root.getStaticEntity('wsrf').putChild('core', core)

    smbinding = SMServiceBinding()
    core.putChild('SubscriptionManagerService', smbinding)

    manager = SubscriptionManager.getInstance(smbinding.__class__, 
        {'wsAddressURI':addr.wsAddressURI})
    
    # Path "wsrf/services": Drop in directory "services"
    dropins = static.File("services")
    dropins.ignoreExt(".rpy")
    dropins.processors = {".rpy": script.ResourceScript}
    root.getStaticEntity('wsrf').putChild('services', dropins)
    
    return root


###########################################################################
# Set up Application
#
###########################################################################

root = Resource()
siteFactory = Site(root)
port = None
kw = {}

root.__class__.server = server = internet.TCPServer(PORT, siteFactory, interface=HOST)
SetUpResourceTree(root)
application = service.Application("WSRF-Container")
server.setServiceParent(application)
